#include <thread>
#include <iostream>

#include "hpp/dds/dds.hpp"
#include "IDL/counter.hpp"


void publishDatas(dds::pub::DataWriter<Idl_Number>& dataWriter)
{
    Idl_Number data;

    int index= 1;
    while(index<11)
    {
        data.number(index);
        std::cerr << "write: "<<index<<std::endl;
        dds::core::InstanceHandle instanceHandler = dataWriter.lookup_instance(data);

        if (instanceHandler.is_nil() == true)
        {
            std::cerr << "register instance " << index << std::endl;
            instanceHandler = dataWriter.register_instance(data);
        }
        dataWriter.write(data,instanceHandler);
//        dataWriter.write(data);
        index++;
    }
}

void disposeDatas(dds::pub::DataWriter<Idl_Number>& dataWriter)
{
    Idl_Number data;
    data.number(7);
    dds::core::InstanceHandle instanceHandler = dataWriter.lookup_instance(data);

    if(instanceHandler.is_nil() == false)
    {
        std::cerr << "unregister Instance 7 " << std::endl;
        dataWriter.unregister_instance(instanceHandler);
    }

    data.number(8);
    instanceHandler = dataWriter.lookup_instance(data);
    if(instanceHandler.is_nil() == false)
    {
        std::cerr << "unregister Instance 8 " << std::endl;
        dataWriter.unregister_instance(instanceHandler);
    }

    data.number(9);
    instanceHandler = dataWriter.lookup_instance(data);
    if(instanceHandler.is_nil() == false)
    {
        std::cerr << "unregister Instance 9 " << std::endl;
        dataWriter.unregister_instance(instanceHandler);
    }

}

int main(int argc, char *argv[])
{
    //****************************** Domain participant ******************************************//
    const short int TYPE_CODE_MAX_SERIALIZED_LENGTH = 20000;
    const int32_t USER_DATA_MAX_LENGHT = 65536;

    // set factory qos
    dds::domain::qos::DomainParticipantFactoryQos factoryQos =
            dds::domain::DomainParticipant::participant_factory_qos();
    factoryQos << dds::core::policy::EntityFactory::ManuallyEnable();
    factoryQos->resource_limits.max_objects_per_thread(4096);
    dds::domain::DomainParticipant::participant_factory_qos(factoryQos);

    // set domain participant qos
    dds::domain::qos::DomainParticipantQos dpQos = dds::core::QosProvider::Default()
            .participant_qos(rti::core::builtin_profiles::qos_lib::baseline());

    rti::core::policy::DomainParticipantResourceLimits resource_limits_qos;
    resource_limits_qos.type_code_max_serialized_length(0);
    resource_limits_qos.reader_user_data_max_length(USER_DATA_MAX_LENGHT);
    resource_limits_qos.writer_user_data_max_length(USER_DATA_MAX_LENGHT);
    resource_limits_qos.type_object_max_serialized_length(TYPE_CODE_MAX_SERIALIZED_LENGTH);
    dpQos << resource_limits_qos;

    std::map<std::string,std::string> dpPropertyMap = {{"dds.transport.UDPv4.builtin.recv_socket_buffer_size", "2100000000"},
                                                       {"dds.transport.UDPv4.builtin.send_socket_buffer_size", "2100000000"},
                                                       {"dds.transport.UDPv4.builtin.parent.message_size_max", "65530"},
                                                       {"dds.transport.shmem.builtin.receive_buffer_size", "204857600"}};

    dpQos << rti::core::policy::Property(dpPropertyMap.begin(),dpPropertyMap.end());

    // Create a DomainParticipant
    dds::domain::DomainParticipant domainParticipant = dds::domain::DomainParticipant(1,dpQos);
    domainParticipant.enable();

    //******************************    create publisher      ******************************************//

    // choose a topicName
    std::string topicName = "MY_IDL";

    // create topic with the domainParticipant, the topicName, the DataType
    dds::topic::Topic<Idl_Number> topic(domainParticipant,topicName);

    auto qosPub =  domainParticipant.default_publisher_qos()
            << dds::core::policy::Partition("defaultPartition");

    // create publisher with created domainParticipant
    dds::pub::Publisher publisher(domainParticipant, qosPub);

    //******************************    create data writer      ******************************************//

    std::map<std::string, std::string> qosDwPropertyMap = {{"dds.builtin_type.*.max_size", "2100000000"},
                                                           {"dds.builtin_type.*.alloc_size", "2100000000"},
                                                           {"dds.data_writer.history.memory_manager.fast_pool.pool_buffer_max_size", "700000000"}};
    dds::pub::qos::DataWriterQos qosDw;
    qosDw << rti::core::policy::Property(qosDwPropertyMap.begin(), qosDwPropertyMap.end());

    qosDw << rti::core::policy::DataWriterProtocol().rtps_reliable_writer(
                              rti::core::policy::RtpsReliableWriterProtocol()
                              .min_send_window_size(dds::core::LENGTH_UNLIMITED)
                              .max_send_window_size(dds::core::LENGTH_UNLIMITED))
          << dds::core::policy::Lifespan(dds::core::Duration(10,0))
          << dds::core::policy::WriterDataLifecycle::AutoDisposeUnregisteredInstances()
          << dds::core::policy::History::KeepLast(1)
          << rti::core::policy::PublishMode::Asynchronous()
          << dds::core::policy::Durability::Persistent()
          << dds::core::policy::Reliability::Reliable();
    qosDw->writer_data_lifecycle->autopurge_disposed_instances_delay(dds::core::Duration(0,0));

    // create DataWriter opject with the publisher, the topic, and the dataWriterQos
    dds::pub::DataWriter<Idl_Number> dataWriter(publisher, topic, qosDw);

    // wait for connecting to TEXT_RECEIVER (if dont, samples will not receive)
    std::this_thread::sleep_for(std::chrono::seconds(5));


    //******************************    publish sample     ******************************************//

    std::this_thread::sleep_for(std::chrono::seconds(10));

    std::cerr << "********** publish ********** "<<std::endl;
    publishDatas(dataWriter);
    std::cerr << "********** unregister **********"<<std::endl;
    disposeDatas(dataWriter);
    std::cerr << "********** publish ********** "<<std::endl;
    publishDatas(dataWriter);
    std::cerr << "********** unregister **********"<<std::endl;
    disposeDatas(dataWriter);
    std::cerr << "********** publish ********** "<<std::endl;
    publishDatas(dataWriter);
    std::cerr << "********** unregister **********"<<std::endl;
    disposeDatas(dataWriter);
    std::this_thread::sleep_for(std::chrono::seconds(100));

}
